"""
this module provides the directive excel.
"""

from typing import List
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.tables import RSTTable
from docutils.statemachine import StringList
from docutils.nodes import table as TableNode
from sphinx.util.docutils import SphinxDirective
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet

from .table import Table
from .span import Span
from .coordinate import Coordinate

def get_data(worksheet: Worksheet) -> List[List[str]]:
    return [[str(cell.value) if cell.value else '' for cell in row] for row in worksheet.rows]

def get_spans(worksheet: Worksheet) -> List[Span]:
    return [
        Span(
            Coordinate(row=cell_range.min_row - 1, column=cell_range.min_col - 1),
            rows=cell_range.max_row - cell_range.min_row + 1,
            columns=cell_range.max_col - cell_range.min_col + 1,
        ) for cell_range in worksheet.merged_cell_ranges  # type: ignore
    ]

class ExcelDirective(RSTTable, SphinxDirective):
    """
    an environment for excel
    """

    option_spec = {
        **RSTTable.option_spec,
        'caption': directives.unchanged,
        'sheet': directives.unchanged,
        'headers': directives.nonnegative_int,
        'no-caption': directives.flag,
    }

    def run(self) -> List[TableNode]:
        """
        render this environment
        """

        file_path, *_ = self.arguments  # type: ignore
        _, file_path = self.env.relfn2path(file_path)
        workbook = load_workbook(file_path)
        default_worksheet_name, *_ = workbook.sheetnames
        headers = self.options.get('headers', 1)
        worksheet_name = self.options.get('sheet', default_worksheet_name)
        caption = self.options.get('caption', worksheet_name)
        no_caption = 'no-caption' in self.options

        worksheet = workbook[worksheet_name]
        table = Table(data=get_data(worksheet), spans=get_spans(worksheet), headers=headers)
        self.arguments = [] if no_caption else [caption]
        self.content = StringList(table.render().splitlines())
        return super().run()



from typing import List
from typing import Dict
from typing import Optional

from .span import Span
from .cell import Cell
from .coordinate import Coordinate

class Spans(dict):
    def add(self, span: Span) -> None:
        self[span.coordinate] = span

    def __getitem__(self, coordinate: Coordinate) -> Span:
        if coordinate in self:
            return super().__getitem__(coordinate)
        return Span(coordinate)

    def find(self, coordinate: Coordinate) -> Optional[Span]:
        for span in self.values():
            if coordinate in span:
                return span
        return None

class Table:
    def __init__(self, data: List[List[str]], spans: List[Span], headers: int = 1):
        self.headers = headers

        self.spans = Spans()
        for span in spans:
            self.spans.add(span)

        self.cells = []
        self.columns = 0
        self.rows = 0
        for row, texts in enumerate(data):
            for column, text in enumerate(texts):
                self.cells.append(Cell(text=text, span=self.spans[Coordinate(row, column)]))
                self.columns = max(self.columns, column + 1)
            self.rows = max(self.rows, row + 1)

    def render(self) -> str:
        column_widths = [0] * self.columns
        row_heights = [0] * self.rows
        for cell in self.cells:
            column_widths[cell.coordinate.column] = max(column_widths[cell.coordinate.column], cell.width)
            row_heights[cell.coordinate.row] = max(row_heights[cell.coordinate.row], cell.height)

        string = ''
        for row in range(self.rows):
            cells = [cell for cell in self.cells if cell.coordinate.row == row]
            rendered_text = [
                cell.render(
                    width=column_widths[column],
                    height=row_heights[row],
                    **self.get_cell_border(cell)
                ).splitlines() for column, cell in enumerate(cells)
            ]

            for line in zip(*rendered_text):
                string += ''.join(line)
                string += '\n'

        return string.strip()

    def get_cell_border(self, cell: Cell) -> Dict[str, str]:
        border = {}

        if cell.coordinate.column == 0:
            # 第一列左右都显示 |
            border['west'] = '|'
            border['east'] = '|'
        else:
            # 其他列只显示右边的 |
            border['west'] = ''
            border['east'] = '|'

        if cell.coordinate.row == 0:
            # 第一行上下都显示 -
            border['north'] = '-'
            border['south'] = '-'
        else:
            # 其他行只显示下面的 -
            border['north'] = ''
            border['south'] = '-'

        if cell.coordinate.row == self.headers - 1:
            # 如果该行是表头的最后一行, 下边框显示 =
            border['south'] = '='

        span = self.spans.find(cell.span.coordinate)
        if span and cell.span.coordinate.column < span.coordinate.column + span.columns - 1:
            border['east'] = ' '

        if span and cell.span.coordinate.row < span.coordinate.row + span.rows - 1:
            border['south'] = ' '

        return border



from .coordinate import Coordinate

class Span:
    def __init__(self, coordinate: Coordinate, rows: int = 1, columns: int = 1):
        self.coordinate = coordinate
        self.rows = rows
        self.columns = columns

    def __repr__(self) -> str:
        return f'<span {self.coordinate} -- ({self.coordinate.row + self.rows - 1}, {self.coordinate.column + self.columns - 1})>'

    def __contains__(self, coordinate: Coordinate) -> bool:
        return self.coordinate.row <= coordinate.row <= self.coordinate.row + self.rows - 1 \
            and self.coordinate.column <= coordinate.column <= self.coordinate.column + self.columns - 1



"""
this is the setup of this package.
"""

from setuptools import setup

from sphinxcontrib.excel import __version__

with open('sphinxcontrib/requirements.txt', 'r', encoding='utf8') as file:
    install_requires = list(map(lambda x: x.strip(), file.readlines()))

with open('readme.md', 'r', encoding='utf8') as file:
    long_description = file.read()

setup(
    name='sphinx-excel',
    version='.'.join(map(str, __version__)),
    author='kinopico',
    author_email='zqmillet@qq.com',
    url='https://github.com/zqmillet/sphinx-excel',
    description='an extension for sphinx to display excel table in sphinx documents',
    packages=['sphinxcontrib.excel'],
    install_requires=install_requires,
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    namespace_packages=["sphinxcontrib"],
)



"""
this is the package sphinxcontrib.console.
"""
__version__ = (1, 0, 1)

from typing import Dict
from typing import Any
from sphinx.application import Sphinx

from .excel import ExcelDirective

def setup(application: Sphinx) -> Dict[str, Any]:
    """
    setup extension.
    """
    application.add_directive('excel', ExcelDirective)
    return {"version": __version__, "parallel_read_safe": True}



from unicodedata import east_asian_width

from .span import Span
from .coordinate import Coordinate

width_map = {
    'F': 2,     # full width
    'H': 1,     # half width
    'W': 2,     # wide
    'Na': 1,    # narrow
    'A': 1,     # ambiguous
    'N': 1,     # neutral
}

def _get_char_display_width(char: str) -> int:
    return width_map.get(east_asian_width(char), 1)

def _get_string_display_width(string: str) -> int:
    return sum(_get_char_display_width(char) for char in string)

class Cell:
    def __init__(self, text: str, span: Span):
        self.text = text
        self.span = span

    def render(self, width: int, height: int, south: str = '-', north: str = '-', west: str = '|', east: str = '|') -> str:
        south_east = '+' if south and east else ''
        south_west = '+' if south and west else ''
        north_east = '+' if north and east else ''
        north_west = '+' if north and west else ''

        if south == ' ' and east == ' ':
            south_east = ' '
        elif south == ' ' and west == ' ':
            south_west = ' '
        elif north == ' ' and east == ' ':
            north_east = ' '
        elif north == ' ' and west == ' ':
            north_west = ' '

        if north:
            string = north_west + north * (width - 2) + north_east + '\n'
        else:
            string = ''

        index = 0
        for line in self.text.splitlines():
            string += west + ' ' + line + ' ' * (width - _get_string_display_width(line) - 4) + ' ' + east + '\n'
            index += 1

        for _ in range(index, height - 2):
            string += west + ' ' * (width - 2) + east + '\n'

        if south:
            return string + south_west + south * (width - 2) + south_east
        return string

    def __repr__(self) -> str:
        return f'<cell {self.span} {repr(self.text)}>'

    @property
    def coordinate(self) -> Coordinate:
        return self.span.coordinate

    @property
    def width(self) -> int:
        return 4 + max((_get_string_display_width(line) for line in self.text.splitlines()), default=0)

    @property
    def height(self) -> int:
        return 2 + len(self.text.splitlines())
